#!/usr/bin/env python3

import numpy as np
from pydamage.optim import optim
from scipy.stats import binom
import collections
from pydamage.utils import sort_dict_by_keys, RMSE, create_ct_cc_dict
from pydamage.vuong import vuong_closeness
from pydamage.likelihood_ratio import LR


def fit_models(ref, model_A, model_B, ct_data, cc_data, ga_data, all_bases, wlen, verbose):
    """Performs model fitting and runs Vuong's closeness test

    Args:
        ref (str): name of referene in alignment file
        model_A (pydamage.models): Pydamage H1 Damage model
        model_B (pydamage.models): Pydamage H0 Null model
        ct_data (list of int): List of positions where CtoT transitions were observed
        ga_data (list of int): List of positions where GtoA transitions were observed
        cc_data (list of int): List of positions where C in ref and query
        all_bases (list of int): List of positions where a base is aligned
        wlen (int): window length
        verbose (bool): verbose mode
    """
    all_bases_pos, all_bases_counts = np.unique(
        np.sort(all_bases), return_counts=True)
    c2t_pos, c2t_counts = np.unique(np.sort(ct_data), return_counts=True)
    g2a_pos, g2a_counts = np.unique(np.sort(ga_data), return_counts=True)
    c2t = dict(zip(c2t_pos, c2t_counts))
    g2a = dict(zip(g2a_pos, g2a_counts))

    # Getting Positions and counts of C2C and C2T
    c2t_dict, c2c_dict = create_ct_cc_dict(ct_data=ct_data,
                                           cc_data=cc_data,
                                           wlen=wlen)

    # Adding zeros at positions where no damage is observed
    for i in all_bases_pos:
        if all_bases_counts[i] > 0:
            if i not in c2t:
                c2t[i] = 0
            if i not in g2a:
                g2a[i] = 0
    c2t = sort_dict_by_keys(c2t)
    g2a = sort_dict_by_keys(g2a)

    xdata = np.array(list(c2t.keys()))
    counts = np.array(list(c2t.values()))

    ydata = list(counts/counts.sum())
    qlen = len(ydata)
    ydata_counts = {i: c for i, c in enumerate(ydata)}
    ctot_out = {f"CtoT-{k}": v for k, v in enumerate(ydata)}

    g2a_counts = np.array(list(g2a.values()))
    y_ga = list(g2a_counts/g2a_counts.sum())
    gtoa_out = {f"GtoA-{k}": v for k, v in enumerate(y_ga)}

    for i in range(qlen):
        if i not in ydata_counts:
            ydata_counts[i] = np.nan
        if f"CtoT-{i}" not in ctot_out:
            ctot_out[f"CtoT-{i}"] = np.nan
        if f"GtoA-{i}" not in gtoa_out:
            gtoa_out[f"GtoA-{i}"] = np.nan

    #################
    # MODEL FITTING #
    #################

    # Only fitting model to interval [0,wlen]
    xdata = xdata[:wlen]
    ydata = ydata[:wlen]

    res = {}
    optim_A, stdev_A = optim(function=model_A.fit,  # damage model
                             parameters=model_A.kwds,
                             xdata=xdata,
                             ydata=ydata,
                             bounds=model_A.bounds)
    if optim_A['pmax'] < optim_A['pmin']:  # making sure that fitting makes sense
        optim_A['pmax'] = optim_A['pmin']

    optim_B, stdev_B = optim(function=model_B.fit,  # null model
                             parameters=model_B.kwds,
                             xdata=xdata,
                             ydata=ydata,
                             bounds=model_B.bounds)

    ##########################
    # LIKELIHOOD CALCULATION #
    ##########################

    # position of sites where C in reference
    c_sites = np.array(list(c2t_dict.keys()))
    # counts of C2C at each site
    c2c_count_per_site = np.array(list(c2c_dict.values()))
    # counts of C2T at each site
    binom_k = np.array(list(c2t_dict.values()))
    # all C per site
    binom_n = c2c_count_per_site + binom_k

    # Likelihood for model A - Damage Model
    binom_p_damage = model_A.fit(c_sites, **optim_A)
    LA = binom.logpmf(k=binom_k, n=binom_n, p=binom_p_damage)

    # Likelihood for model B - Null Model
    binom_p_null = model_B.fit(c_sites, **optim_B)
    LB = binom.logpmf(k=binom_k, n=binom_n, p=binom_p_null)

    # Difference of number of paramters between model A and model B
    pdiff = len(model_A.kwds) - len(model_B.kwds)

    ################
    # LR TEST #
    ################

    LR_lambda, pval = LR(L0=LB, L1=LA, df=pdiff)

    res.update(ydata_counts)
    res.update(ctot_out)
    res.update(gtoa_out)
    res.update(optim_A)
    res.update(stdev_A)
    res.update(optim_B)
    res.update(stdev_B)
    res.update({'pvalue': pval})
    res.update({'base_cov': all_bases_counts})
    res.update({'model_params': list(optim_A.values()) +
                list(optim_B.values())+list(stdev_A.values())+list(stdev_B.values())})
    res['qlen'] = qlen
    res['residuals'] = ydata - model_A.fit(x=xdata, **optim_A)
    res['RMSE'] = RMSE(res['residuals'])
    res['wlen'] = wlen
    return(res)



import numpy as np
import os
import sys
import shutil
from statsmodels.stats.multitest import multipletests
import pandas as pd


def check_extension(filename):
    """Check alignment file format to give correct open mode

    Args:
        filename (str): Path to alignment file

    Returns:
        str: opening mode

    Raises:
        Exception: Extension not supported
    """
    extension = filename.split(".")[-1]
    modes = {'bam': 'rb', 'sam': 'r', 'cram': 'rc'}
    try:
        return(modes[extension])
    except KeyError:
        raise Exception(f"{extension} file extension not supported")


def makedir(dirpath, confirm=True, force=False):
    """Make directory if user confirms overwritting

    Args:
        dirpath (str): Path to directory
        confirm (bool, optional): Ask the user to confirm. Defaults to True.
        force (bool, optional): Always create directory. Defaults to False.
    """
    if os.path.exists(dirpath):
        if confirm and force == False:
            print(
                f"Result directory, {dirpath}, already exists, it will be overwritten")
            if input('Do You Want To Continue? (y|n) ').lower() != 'y':
                sys.exit()
        shutil.rmtree(dirpath)

    os.makedirs(dirpath)


def pandas_processing(res_dict, outdir):
    """Performs Pandas processing of Pydamage results

    Args:
        res_dict (dict): Result dictionary of Vuong's closeness test
        outdir (str): Path to output directory
    """
    df = pd.DataFrame(res_dict)
    if len(res_dict) == 0:
        return(df)
    qvalues = pd.Series(multipletests(df['pvalue'].dropna(), method='fdr_bh')[
                        1], index=df['pvalue'].dropna().index, name='qvalue')
    df = df.merge(qvalues, left_index=True, right_index=True, how='outer')
    df = df[['p0', 'p0_stdev',
             'p', 'p_stdev',
             'pmin', 'pmin_stdev',
             'pmax', 'pmax_stdev',
             'pvalue',
             'qvalue',
             'RMSE',
             'reference',
             'nb_reads_aligned',
             'coverage'] +
            [f"CtoT-{i}" for i in range(df['qlen'].max())] +
            [f"GtoA-{i}" for i in range(df['qlen'].max())]]
    df.rename(columns={'p0': 'null_model_p0',
                       'p0_stdev': 'null_model_p0_stdev',
                       'p': 'damage_model_p',
                       'p_stdev': 'damage_model_p_stdev',
                       'pmin': 'damage_model_pmin',
                       'pmin_stdev': 'damage_model_pmin_stdev',
                       'pmax': 'damage_model_pmax',
                       'pmax_stdev': 'damage_model_pmax_stdev'},
              inplace=True)
    df.sort_values(by=['qvalue'], inplace=True)
    df.set_index("reference", inplace=True)
    df.dropna(axis=1, how='all', inplace=True)

    df.to_csv(f"{outdir}/pydamage_results.csv")
    return(df)


def sort_dict_by_keys(adict):
    """Sort dictonary by keys

    Args:
        adict (dict): dictorary to sort
    Returns:
        dict: dictionary sorted on keys
    """
    res = {}
    for k in sorted(adict.keys()):
        res[k] = adict[k]
    return(res)


def RMSE(residuals):
    """Computes Root Mean Square Error

    Args:
        residuals (np.array(float)): Array of residuals
    Returns:
        float: RMSE
    """
    return(np.sqrt(np.mean(residuals**2)))


def create_ct_cc_dict(ct_data, cc_data, wlen):
    """Creates C bases positions dictionnary

    For C->T transitions
    For C in reference and in query

    Args:
        ct_data (list of int): List of positions where CtoT transitions were observed
        cc_data (list of int): List of positions where C in ref and query
    Returns:
        dict{int:int}: {position: number of CtoT transitions at position}
        dict{int:int}: {position: number of C match at position}
    """
    c2t_pos, c2t_counts = np.unique(np.sort(ct_data), return_counts=True)
    c2c_pos, c2c_counts = np.unique(np.sort(cc_data), return_counts=True)
    c2t = dict(zip(c2t_pos, c2t_counts))
    c2c = dict(zip(c2c_pos, c2c_counts))
    c2t_dict = {}
    c2c_dict = {}

    for i in range(wlen):
        if i in c2c or i in c2t:
            if i in c2t:
                c2t_dict[i] = c2t[i]
            else:
                c2t_dict[i] = 0
            if i in c2c:
                c2c_dict[i] = c2c[i]
            else:
                c2c_dict[i] = 0

    return(c2t_dict, c2c_dict)



import matplotlib.pyplot as plt
from pydamage.models import damage_model, null_model
from statsmodels.nonparametric.smoothers_lowess import lowess
from pydamage import utils
import numpy as np
from os import makedirs
from scipy.stats import probplot


def damageplot(damage_dict, outdir):
    """Draw pydamage plots

    Args:
        damage_dict(dict): pydamage result dictionary
        qlen(int): query length
        outdir(str): Pydamage result directory
    """
    x = np.array(range(damage_dict['wlen']))
    qlen = np.array(range(damage_dict['qlen']))
    y = np.array([damage_dict[i] for i in x])
    c2t = np.array([damage_dict[f"CtoT-{i}"] for i in qlen])
    g2a = np.array([damage_dict[f"GtoA-{i}"] for i in qlen])
    p0 = damage_dict['p0']
    p0_stdev = damage_dict['p0_stdev']
    p = damage_dict['p']
    pmin = damage_dict['pmin']
    pmin_stdev = damage_dict['pmin_stdev']
    pmax = damage_dict['pmax']
    pmax_stdev = damage_dict['pmax_stdev']
    contig = damage_dict['reference']
    pvalue = damage_dict['pvalue']
    coverage = damage_dict['coverage']
    residuals = damage_dict['residuals']
    rmse = damage_dict['RMSE']
    plotdir = outdir

    if pvalue < 0.001:
        rpval = "<0.001"
    else:
        rpval = f"={round(pvalue,3)}"

    m_null = null_model()
    p0_low = max(m_null.bounds[0][0], p0 - 2*p0_stdev)
    p0_high = min(m_null.bounds[1][0], p0 + 2*p0_stdev)
    y_unif = m_null.fit(x, p0)
    y_unif_low = np.maximum(
        np.zeros(y_unif.shape[0]), m_null.fit(x, p0_low))
    y_unif_high = np.minimum(
        np.ones(y_unif.shape[0]), m_null.fit(x, p0_high))

    geom = damage_model()
    geom_pmin_low = max(geom.bounds[0][1], pmin - 2*pmin_stdev)
    geom_pmin_high = min(geom.bounds[1][1], pmin + 2*pmin_stdev)
    geom_pmax_low = max(geom.bounds[0][2], pmax - 2*pmax_stdev)
    geom_pmax_high = min(geom.bounds[1][2], pmax + 2*pmax_stdev)

    y_geom = geom.fit(x, p, pmin, pmax)
    y_geom_low = np.maximum(np.zeros(y_geom.shape[0]), geom.fit(
        x, p, geom_pmin_low, geom_pmax_low))
    y_geom_high = np.minimum(np.ones(y_geom.shape[0]), geom.fit(
        x, p, geom_pmin_high, geom_pmax_high))

    plt.xticks(rotation=45, fontsize=8)

    fig, ax = plt.subplots()

    ax.plot(x, y_unif,
            linewidth=2.5,
            color='DarkOliveGreen',
            alpha=0.8,
            label='Null model')

    ax.fill_between(x, y_unif_low, y_unif_high,
                    color='DarkOliveGreen',
                    alpha=0.1,
                    label='Null Model CI (2 sigma)')

    ax.plot(x, y_geom,
            linewidth=2.5,
            color='#D7880F',
            alpha=0.8,
            label='Damage model')

    ax.fill_between(x, y_geom_low, y_geom_high,
                    color='#D7880F',
                    alpha=0.1,
                    label='Damage Model CI (2 sigma)')

    ax.plot(qlen, g2a,
            color='#236cf5',
            alpha=0.1,
            label='G to A transitions')

    ax.plot(qlen, c2t,
            color='#bd0d45',
            alpha=0.2,
            label='C to T transitions')

    ax.set_xlabel("Base from 5'", fontsize=10)
    ax.set_ylabel("Substitution frequency", fontsize=10)
    ax.xaxis.set_ticks(np.arange(qlen[0], qlen[-1], 5))
    ax.set_xticklabels(ax.get_xticks(), rotation=45, fontsize=6)
    ax.set_title(f"coverage: {round(coverage,2)} - pvalue{rpval}", fontsize=8)
    ax.legend(fontsize=8)
    # ax.set_title(f"coverage: {round(coverage,2)} | pvalue{rpval}", fontsize=8)

    left, bottom, width, height = [0.65, 0.3, 0.2, 0.2]
    ax2 = fig.add_axes([left, bottom, width, height])

    fitted = x
    smoothed = lowess(residuals, fitted)
    ax2.scatter(fitted, residuals, marker=".", color='black')
    ax2.plot(smoothed[:, 0], smoothed[:, 1], color='r')
    ax2.plot([min(fitted), max(fitted)], [0, 0],
             color='k', linestyle=':', alpha=.3)
    ax2.set_ylabel('Residuals', fontsize=6)
    ax2.set_xlabel('Fitted Values', fontsize=6)
    ax2.set_title(f"Residuals vs. Fitted\nRMSE={round(rmse, 3)}", fontsize=6)
    ax2.xaxis.set_ticks(np.arange(fitted[0], fitted[-1], 5))
    ax2.set_xticklabels([int(i) for i in ax2.get_xticks()],
                        fontsize=6, rotation=45)
    ax2.set_yticklabels([round(i, 3) for i in ax2.get_yticks()], fontsize=6)

    fig.suptitle(contig, fontsize=12, y=0.95)

    fig.savefig(f"{plotdir}/{contig}.png", dpi=200)
